#!/usr/bin/perl

use strict;
use warnings;
no warnings 'uninitialized', 'once';
use LWP::UserAgent;

$|++;

our ($VERSION) = "1.01";
our ($PROGRAM) = 'PCCC Raptor FAST';
our ($COPYRIGHT) = 'Copyright (C) 2014 Peregrine Computer Consultants Corporation. All rights reserved.';
our ($THANKS) = 'Thanks to Brian Landers <brian@packetslave.com> for his inspirational work on integrating sendmail with LDAP and to Tony Torres for his help in the initial development and live testing for this program.';

#EXCHANGE CONSTANTS
our $ACCOUNTDISABLE = 2;
our $LOCKOUT = 16;

use Net::LDAP;
use Net::LDAP::Control::Paged;
use Net::LDAP::Constant qw( LDAP_CONTROL_PAGED );

my (%FORM, $FORM);
$FORM = \%FORM;

print "$PROGRAM v$VERSION\n$COPYRIGHT\n\n$THANKS\n\n";

if (-e "config") {
  print "Loading Configuration... \n";
  require "config";
  if ($main::token eq '') {
    print "Error.  You must have a valid Raptor FAST Token defined in your config file!\nExiting.\n\n";
    exit;
  }
} else {
  print "Error Loading Configuration! You must copy config.dist to config and edit it for your installation.\n";
  exit;
}

&main($FORM, bind=>$main::bind, password=>$main::password, base=>$main::base, servers=>\@main::servers, filter=>$main::filter, token=>$main::token, verbose=>$main::VERBOSE);
exit;


sub main {
  my ($FORM, %params) = @_;
  my ($ldap, %FORM, $rv, $status, $msg, $key, $output, $i);

  #CONNECT/LOGIN TO THE LDAP DATABASE
  ($rv, $status, $ldap) = &ldap_connect_and_bind(\%FORM, servers=>$params{'servers'}, bind=>$params{'bind'}, password=>$params{'password'});
  if ($rv < 1) {
    print "Fatal $status\nExiting...\n\n";
    exit;
  }

  #SEARCH THE DIRECTORY
  our %email_addresses;
  ($rv, $status) = &ldap_search($FORM, ldap=>$ldap, filter=>$params{'filter'}, base=>$params{'base'});
  #print "DEBUG: $rv - $status\n";

  if ($rv) {
    #DISPLAY THE RESULTS
    $i=0;
    foreach $key (keys %email_addresses) {
      $i++;
      $output .= "$key\n";
    }
    print "DEBUG: $i email addresses found\n $output\n" if ($params{'VERBOSE'});
  }

  #UNBIND/LOGOUT AND DISCONNECT FROM THE SERVER
  ($rv, $status) = &ldap_unbind_and_disconnect($FORM, ldap=>$ldap);
  print "$status\n";

  #REPORT THE RESULTS TO THE RAPTOR SERVER
  &submit_emails_to_raptor($FORM, emails=>$output, token=>$params{'token'});

  #TEST FOR SCHEDULING PURPOSES

}

sub submit_emails_to_raptor {
  my ($FORM, %params) = @_;
  my ($user_agent, $request, $result);

  #REPORT THE RESULTS TO THE RAPTOR SERVER
  $user_agent = new LWP::UserAgent;
  $request = new HTTP::Request("POST", "https://www.pccc.com/raptor_FAST_submit.cgim");

  $request->content_type("application/x-www-form-urlencoded");

  #URL ENCODE THE INFORMATION
  $params{'emails'} =~ s/([^a-zA-Z0-9_.-])/uc sprintf("%%%02x",ord($1))/eg;

  $request->content("token=$params{'token'}&emails=$params{'emails'}");
  $result = $user_agent->request($request);
  if ($result->is_error) { 
    print "Error submitting LDAP information to Raptor.\n";
  } else {
    print "Raptor submission succesful.\n";
  }
}

#MASTER LDAP OBJECT HANDLER
sub handle_ldap_object {
  my ($msg, $data) = @_;

  #print "DEBUG: handle_ldap_object - $msg - $data\n";
  if ($data) {
    if (uc(ref($data)) eq 'NET::LDAP::ENTRY') {
      return handle_ldap_entry($msg, $data);
    } elsif (uc(ref($data)) eq 'NET::LDAP::REFERENCE') {
      return handle_ldap_reference($msg, $data);
    } else {
      #UNKNOWN
    }
  }
  return;
}

sub handle_ldap_entry {
  # Net::LDAP::Entry HANDLER
  # This is the key record handler.   
  # Ignore accounts that are disabled or locked out and extract the SMTP email 
  # addresses from the record and output only the SMTP ones we haven't seen before.
  my ($msg, $data) = @_;
  my (@emails, $email);
 
  #print "DEBUG: ".$data->get_value('proxyAddresses')."\n\n";

  # Extract the SMTP emails from the LDAP data and check each
  @emails = grep { /^smtp:/i } $data->get_value("proxyAddresses");
 
  # If we found any, strip off the SMTP: identifier and print them out 
  foreach $email (@emails) {
    if (int($data->get_value( "userAccountControl" )) & $ACCOUNTDISABLE) {
      #print "DISABLED: ";
    } elsif (int($data->get_value( "userAccountControl" )) & $LOCKOUT) {
      #print "LOCKED OUT: ";
    } else {
      $email =~ s/^smtp:(.+)$/$1/i;
      $main::email_addresses{$email}++;
    }
  }
}

sub handle_ldap_reference {
  my ($msg, $data) = @_; 
  my ($sub_data);
  
  foreach $sub_data ($data->references()) {
    return handle_ldap_object($msg, $sub_data);
  }
}

sub ldap_search {
  my ($FORM, %params) = @_;
  my ($msg, $entry, $page, @search_args, $resp, $cookie);

  $params{'page_size'} ||= 100;

  if (ref($params{'ldap'}) !~ /^Net::LDAP/i) {
    return (0, 'Invalid LDAP Handle Received');
  }

  # Exchange has a maximum recordset size of 1000 records by default.  Use paging to get the full list.
  $page = Net::LDAP::Control::Paged->new( size => $params{'page_size'} );

  print "Running a Search on a base of: [$params{'base'}] with the following filter: [$params{'filter'}]\n";
  @search_args = ( attrs=>['proxyAddresses','userAccountControl'],callback=>\&handle_ldap_object, control=>[$page],
                   scope=>'subtree', filter=>$params{'filter'}, base=>$params{'base'});

  #DEBUG SEARCH
  if ($params{'debug'}) {
    $msg = $params{'ldap'}->search(scope=>'subtree', filter=>$params{'filter'}, base=>$params{'base'});
    foreach $entry ($msg->all_entries) { 
      print "DEBUG: ";
      print $entry->dump; 
      print "\n\n";
    }
    return (1, 'DEBUG: Search Results Returned');
  } else {
    #SEARCH AND PARSE THE RESULTS NORMALLY
    while(1) {
      # Perform search
      $msg = $params{'ldap'}->search(@search_args);

      # Only continue on LDAP_SUCCESS
      if ($msg->code) { last };

      # Get cookie from paged control
      ($resp) = $msg->control( LDAP_CONTROL_PAGED ) or last;

      $cookie = $resp->cookie or last;

      # Set cookie in paged control
      $page->cookie($cookie);
    }
    if ($cookie) {
      # We had an abnormal exit, so let the server know we do not want any more
      $page->cookie($cookie);
      $page->size(0);
      $params{'ldap'}->search( @search_args );
    }
    return (1, 'Search Results Returned');
  }
}

sub print_ldap_error {
  print &get_ldap_error(@_);
}

sub get_ldap_error {
  my ($FORM, %params) = @_;
  my ($output);

  if (ref($params{'msg'}) =~ /^Net::LDAP/i) {
    if ($params{'msg'}->code()) {
       $output = "Error: LDAP Code:".$params{'msg'}->code()." / Error: ".$params{'msg'}->error." / Error Text:".$params{'msg'}->error_text()."\n";
    }
  } else {
    $output = "Error: Invalid LDAP Message Handle\n";
  }

  return $output;
}

sub ldap_unbind_and_disconnect {
  my ($FORM, %params) = @_;

  if (uc(ref($params{'ldap'})) eq 'NET::LDAP') {
    $params{'ldap'}->unbind();
    $params{'ldap'}->disconnect();
    undef $params{'ldap'};
    print "DEBUG: Disconnected.\n" if ($params{'debug'});
    return (0, 'Unbound and Disconnected from LDAP server.');
  } else {
    return (1, 'Invalid LDAP handle passed for Unbind / Disconnect.');
  }
}

sub ldap_connect_and_bind {
  my ($FORM, %params) = @_;
  my ($i, $ldap, $msg);

  #Connect to the server - try each one by one until we succeed
  $i = 0;
  while ($i < scalar(@{$params{'servers'}}) && !$ldap) {
    print "Trying to Connect to LDAP Server #$i: $params{'servers'}->[$i]\n";
    $ldap = Net::LDAP->new( $params{'servers'}->[$i] );
  }

  #CONNECT STATUS
  if (uc(ref($ldap)) eq 'NET::LDAP') {
    print "Connected!\n";

    print "DEBUG: Bind is: $params{'bind'} / $params{'password'}\n" if ($params{'debug'});

    #BIND (LOGIN) TO THE LDAP SERVER
    $msg = $ldap->bind( dn=>$params{'bind'}, password=>$params{'password'} );

    # If we can't bind, we can't continue
    if( $msg->code() ) {
      return (0, &get_ldap_error($FORM, msg=>$msg));
    } else {
      print "DEBUG: Login Successful!\n" if ($params{'debug'});
      return (1, "Login Successful!\n", $ldap);
    }

  } else {
    return (0, 'Unable to connect to any LDAP servers!');
  }
}

sub test_perl {
  #TEST FOR PERL FUNCTIONALITY
  my ($i);
  for ($i = 0; $i < 1000000; $i++) {
    print "Hello World!";
  }
}
